/*
 * Author: Radovan Cerveny
 */

#include "MeasurementProvisioner.hpp"
#include "exception/CommonException.h"
#include "sax/SaxParseInterface.h"
#include <wait.h>
#include <iostream>

#include "MeasurementProvisionerUtils.hpp"

using namespace std;

namespace provisioner {

MeasurementProvisionerResults MeasurementProvisioner::runConfiguration ( const MeasurementProvisionerConfiguration & cfg ) {

	prepareEnvironment ( cfg );

	MPInputData mpiData ( cfg );

	MeasurementProvisionerResults mprResults;

	for ( const MPSubstitutionMap & substitutionMap : mpiData.getSubstitutionMaps ( ) ) {
		MPRResults mprr;

		for ( const auto & mpsmPair : substitutionMap )
			mprr.inputs.push_back ( { mpsmPair.first, mpiData.getMPIDHandle ( mpsmPair.second ).alias } );

		for ( const MPPipeline & pipeline : mpiData.getSubstitutionPipelines ( ) )
			mprr.pipelineResults.push_back ( runPipeline ( pipeline, substitutionMap ) );

		mprResults.results.push_back ( std::move ( mprr ) );
	}

	return mprResults;
}

void MeasurementProvisioner::prepareEnvironment ( const MeasurementProvisionerConfiguration & cfg ) {
	list < string > binariesDirectoryExpansion = MPUtils::shellExpand ( cfg.binariesDirectory );

	if ( binariesDirectoryExpansion.size ( ) != 1 )
		throw::exception::CommonException ( "MeasurementProvisioner: binaries directory: \"" + cfg.binariesDirectory + "\"expansion failed" );

	if ( chdir ( binariesDirectoryExpansion.begin ( )->c_str ( ) ) != 0 )
		throw::exception::CommonException ( "MeasurementProvisioner: chdir to binaries directory: \"" + cfg.binariesDirectory + "\" failed" );
}

MPRPipelineResults MeasurementProvisioner::runPipeline ( const MPPipeline & pipeline, const MPSubstitutionMap & substitutionMap ) {

	int measurementsTmpfileFd = open ( "/tmp", O_RDWR | O_TMPFILE, S_IRUSR | S_IWUSR );
	int measurementsFd = measurementsTmpfileFd;

	if ( measurementsTmpfileFd == -1 )
		throw::exception::CommonException ( "MeasurementProvisioner: Cannot make measurements tmp file" );

	 // reserve fd 5 for measurements
	if ( measurementsFd != 5 )
		measurementsFd = dup2 ( measurementsFd, 5 );

	if ( measurementsFd == -1 )
		throw::exception::CommonException ( "MeasurementProvisioner: Cannot open measurements file descriptor" );

	int inputTmpfileFd	= open ( "/tmp", O_RDWR | O_TMPFILE, S_IRUSR | S_IWUSR );
	int outputTmpfileFd = open ( "/tmp", O_RDWR | O_TMPFILE, S_IRUSR | S_IWUSR );
	int errorTmpfileFd	= open ( "/tmp", O_RDWR | O_TMPFILE, S_IRUSR | S_IWUSR );

	if ( ( inputTmpfileFd == -1 ) || ( outputTmpfileFd == -1 ) || ( errorTmpfileFd == -1 ) )
		throw::exception::CommonException ( "MeasurementProvisioner: Cannot make io tmp files" );

	 // store handles to stdin, stdout, stderr, we will need to restore them later
	int curStdinFd	= dup ( 0 );
	int curStdoutFd = dup ( 1 );
	int curStderrFd = dup ( 2 );
	ofdstream ofdlog ( curStderrFd );

	close ( 0 );
	close ( 1 );
	close ( 2 );

	MPRPipelineResults pipelineResults { };

	for ( MPPipeline::const_iterator mppcIter = pipeline.cbegin ( ); mppcIter != pipeline.cend ( ); ++mppcIter ) {
		if ( mppcIter == pipeline.cbegin ( ) ) {
			dup ( inputTmpfileFd );
			dup ( outputTmpfileFd );
			dup ( errorTmpfileFd );
		} else {
			 // swap stdin and stdout for next command
			close ( 0 );
			close ( 1 );
			dup ( outputTmpfileFd );
			dup ( inputTmpfileFd );

			 // rewind previous stdout to the beginning, so the next command can read the whole file as input
			lseek ( 0, SEEK_SET, 0 );

			 // destroy the contents of previous stdin and rewind to the beginning
			ftruncate ( 1, 0 );
			lseek ( 1, SEEK_SET, 0 );
		}

		string command = mppcIter->substitute ( substitutionMap ) + " -m";

		ofdlog << command << endl;

		 // leave the execution to the system shell
		int status = system ( command.c_str ( ) );

		ofdlog << WEXITSTATUS ( status ) << endl;

		if ( WEXITSTATUS ( status ) != 0 ) {
			ifdstream errorIfd ( errorTmpfileFd );
			stringstream buffer;
			buffer << errorIfd.rdbuf ( );
			pipelineResults.pipelineStatus.exitCode = WEXITSTATUS ( status );
			pipelineResults.pipelineStatus.errorOrigin = mppcIter->getRawCommand ( );
			pipelineResults.pipelineStatus.errorValue  = buffer.str ( );
			break;
		}

		 // since we are at the end of file after the output, we need to rewind
		lseek ( measurementsFd, SEEK_SET, 0 );
		ifdstream ifd ( measurementsFd );

		deque < sax::Token > tokens;
		sax::SaxParseInterface::parseStream ( ifd, tokens );

		measurements::MeasurementResults measurementResults = measurements::MeasurementXmlFactory::parseMeasurementResults ( tokens );

		ofdlog << measurements::MeasurementFormat::TREE << measurementResults << endl;

		pipelineResults.commandResults.push_back ( { mppcIter->getRawCommand ( ), std::move ( measurementResults ) } );

		 // destroy the contents and rewind for another commands measurements
		ftruncate ( measurementsFd, 0 );
		lseek ( measurementsFd, SEEK_SET, 0 );
	}

	 // close tmp files, which results in their deletion
	if ( measurementsFd != measurementsTmpfileFd )
		close ( measurementsTmpfileFd );

	close ( measurementsFd );
	close ( inputTmpfileFd );
	close ( outputTmpfileFd );
	close ( errorTmpfileFd );

	 // restore our main stdin, stdout, stderr
	close ( 0 );
	close ( 1 );
	close ( 2 );

	dup ( curStdinFd );
	dup ( curStdoutFd );
	dup ( curStderrFd );

	return pipelineResults;
}

}
