/*
 * Author: Radovan Cerveny
 */

#include "sax/FromXMLParserHelper.h"
#include "MeasurementProvisionerConfigurationXml.hpp"
#include "exception/CommonException.h"

using namespace std;
using namespace sax;

namespace provisioner {

MeasurementProvisionerConfiguration MeasurementProvisionerConfigurationXml::parse ( deque < Token > & tokens ) {
	MeasurementProvisionerConfiguration configuration;

	deque < Token >::iterator iter = tokens.begin ( );

	if ( iter == tokens.end ( ) ) throw::exception::CommonException ( "Empty tokens list" );

	parseMeasurementProvisioner ( configuration, iter );

	if ( iter != tokens.end ( ) ) throw::exception::CommonException ( "Unexpeted tokens at the end of the xml" );

	return configuration;
}

void MeasurementProvisionerConfigurationXml::parseMeasurementProvisioner ( MeasurementProvisionerConfiguration & mpc, std::deque < sax::Token >::iterator & iter ) {
	FromXMLParserHelper::popToken ( iter, Token::TokenType::START_ELEMENT, MEASUREMENT_PROVISIONER_TAG );

	for ( ; ; ) {
		if ( !FromXMLParserHelper::isTokenType ( iter, Token::TokenType::START_ELEMENT ) ) break;

		string data = FromXMLParserHelper::getTokenData ( iter, Token::TokenType::START_ELEMENT );

		if ( data == BINARIES_DIRECTORY_TAG ) {
			FromXMLParserHelper::popToken ( iter, Token::TokenType::START_ELEMENT, BINARIES_DIRECTORY_TAG );
			mpc.binariesDirectory = FromXMLParserHelper::popTokenData ( iter, Token::TokenType::CHARACTER );
			FromXMLParserHelper::popToken ( iter, Token::TokenType::END_ELEMENT, BINARIES_DIRECTORY_TAG );
		} else if ( data == MEASUREMENT_PROVISION_TAG ) {
			parseMeasurementProvision ( mpc, iter );
		} else {
			break;
		}
	}

	FromXMLParserHelper::popToken ( iter, Token::TokenType::END_ELEMENT, MEASUREMENT_PROVISIONER_TAG );
}

void MeasurementProvisionerConfigurationXml::parseMeasurementProvision ( MeasurementProvisionerConfiguration & mpc, std::deque < sax::Token >::iterator & iter ) {
	FromXMLParserHelper::popToken ( iter, Token::TokenType::START_ELEMENT, MEASUREMENT_PROVISION_TAG );

	for ( ; ; ) {
		if ( !FromXMLParserHelper::isTokenType ( iter, Token::TokenType::START_ELEMENT ) ) break;

		string data = FromXMLParserHelper::getTokenData ( iter, Token::TokenType::START_ELEMENT );

		if ( data == INPUT_DATA_TAG )
			parseInputData ( mpc, iter );
		else if ( data == PIPELINES_TAG )
			parsePipelines ( mpc, iter );
		else if ( data == MEASUREMENT_CONFIGURATION_TAG )
			parseMeasurementConfiguration ( mpc, iter );
		else
			break;
	}

	FromXMLParserHelper::popToken ( iter, Token::TokenType::END_ELEMENT, MEASUREMENT_PROVISION_TAG );
}

void MeasurementProvisionerConfigurationXml::parseInputData ( MeasurementProvisionerConfiguration & mpc, std::deque < sax::Token >::iterator & iter ) {
	FromXMLParserHelper::popToken ( iter, Token::TokenType::START_ELEMENT, INPUT_DATA_TAG );

	for ( ; ; ) {
		if ( !FromXMLParserHelper::isTokenType ( iter, Token::TokenType::START_ELEMENT ) ) break;

		string data = FromXMLParserHelper::getTokenData ( iter, Token::TokenType::START_ELEMENT );

		if ( data == INPUT_DATA_BATCH_TAG )
			parseInputBatch ( mpc, iter );
		else
			break;
	}

	FromXMLParserHelper::popToken ( iter, Token::TokenType::END_ELEMENT, INPUT_DATA_TAG );
}

void MeasurementProvisionerConfigurationXml::parseInputBatch ( MeasurementProvisionerConfiguration & mpc, std::deque < sax::Token >::iterator & iter ) {
	FromXMLParserHelper::popToken ( iter, Token::TokenType::START_ELEMENT, INPUT_DATA_BATCH_TAG );

	MPCInputBatch mpcib;

	mpcib.forcedDependencyBatch = false;

	if ( FromXMLParserHelper::isTokenType ( iter, Token::TokenType::START_ATTRIBUTE ) ) {
		string attrName = FromXMLParserHelper::getTokenData ( iter, Token::TokenType::START_ATTRIBUTE );

		if ( attrName == INPUT_DATA_BATCH_ATTR_DEPENDENCY_NAME ) {
			FromXMLParserHelper::popToken ( iter, Token::TokenType::START_ATTRIBUTE, attrName );
			string attrVal = FromXMLParserHelper::popTokenData ( iter, Token::TokenType::CHARACTER );

			if ( attrVal == "true" )
				mpcib.forcedDependencyBatch = true;

			FromXMLParserHelper::popToken ( iter, Token::TokenType::END_ATTRIBUTE, attrName );
		}
	}

	for ( ; ; ) {
		if ( !FromXMLParserHelper::isTokenType ( iter, Token::TokenType::START_ELEMENT ) ) break;

		string data = FromXMLParserHelper::getTokenData ( iter, Token::TokenType::START_ELEMENT );

		mpcib.batch.push_back ( parseInputDatum ( data, iter ) );
	}

	mpc.inputBatches.push_back ( std::move ( mpcib ) );

	FromXMLParserHelper::popToken ( iter, Token::TokenType::END_ELEMENT, INPUT_DATA_BATCH_TAG );
}

MPCInputDatum MeasurementProvisionerConfigurationXml::parseInputDatum ( const std::string & tag, std::deque < sax::Token >::iterator & iter ) {
	MPCInputDatum mpcid;

	mpcid.id = -1;
	mpcid.count = 1;

	if ( tag == INPUT_DATA_FILE_TAG )
		mpcid.type = MPCInputDatum::Type::FILE;
	else if ( tag == INPUT_DATA_GENERATOR_TAG )
		mpcid.type = MPCInputDatum::Type::GENERATOR;
	else
		throw::exception::CommonException ( "XML: Unknown MPCInputDatum type" );

	FromXMLParserHelper::popToken ( iter, Token::TokenType::START_ELEMENT, tag );

	bool foundId = false;

	for ( ; ; ) {
		if ( !FromXMLParserHelper::isTokenType ( iter, Token::TokenType::START_ATTRIBUTE ) ) break;

		string attrName = FromXMLParserHelper::getTokenData ( iter, Token::TokenType::START_ATTRIBUTE );
		FromXMLParserHelper::popToken ( iter, Token::TokenType::START_ATTRIBUTE, attrName );
		string attrVal = FromXMLParserHelper::popTokenData ( iter, Token::TokenType::CHARACTER );
		FromXMLParserHelper::popToken ( iter, Token::TokenType::END_ATTRIBUTE, attrName );

		if ( attrName == INPUT_DATA_ATTR_ID_NAME ) {
			mpcid.id = std::stoi ( attrVal );
			foundId	 = true;
		} else if ( attrName == INPUT_DATA_ATTR_COUNT_NAME ) {
			mpcid.count = std::stoi ( attrVal );
		} else {
			throw::exception::CommonException ( "XML: MPCInputDatum unknown attr: " + attrName );
		}
	}

	if ( !foundId )
		throw::exception::CommonException ( "XML: MPCInputDatum id attr was not specified" );

	mpcid.value = FromXMLParserHelper::popTokenData ( iter, Token::TokenType::CHARACTER );

	FromXMLParserHelper::popToken ( iter, Token::TokenType::END_ELEMENT, tag );
	return mpcid;
}

void MeasurementProvisionerConfigurationXml::parsePipelines ( MeasurementProvisionerConfiguration & mpc, std::deque < sax::Token >::iterator & iter ) {
	FromXMLParserHelper::popToken ( iter, Token::TokenType::START_ELEMENT, PIPELINES_TAG );

	for ( ; ; ) {
		if ( !FromXMLParserHelper::isTokenType ( iter, Token::TokenType::START_ELEMENT ) ) break;

		string data = FromXMLParserHelper::getTokenData ( iter, Token::TokenType::START_ELEMENT );

		if ( data == PIPELINE_TAG ) {
			mpc.pipelines.push_back ( { } );
			parsePipeline ( mpc.pipelines.back ( ), iter );
		} else {
			break;
		}
	}

	FromXMLParserHelper::popToken ( iter, Token::TokenType::END_ELEMENT, PIPELINES_TAG );
}

void MeasurementProvisionerConfigurationXml::parsePipeline ( std::vector < std::string > & pipeline, std::deque < sax::Token >::iterator & iter ) {
	FromXMLParserHelper::popToken ( iter, Token::TokenType::START_ELEMENT, PIPELINE_TAG );

	for ( ; ; ) {
		if ( !FromXMLParserHelper::isTokenType ( iter, Token::TokenType::START_ELEMENT ) ) break;

		string data = FromXMLParserHelper::getTokenData ( iter, Token::TokenType::START_ELEMENT );

		if ( data == COMMAND_TAG ) {
			FromXMLParserHelper::popToken ( iter, Token::TokenType::START_ELEMENT, COMMAND_TAG );
			pipeline.emplace_back ( FromXMLParserHelper::popTokenData ( iter, Token::TokenType::CHARACTER ) );
			FromXMLParserHelper::popToken ( iter, Token::TokenType::END_ELEMENT, COMMAND_TAG );
		} else {
			break;
		}
	}

	FromXMLParserHelper::popToken ( iter, Token::TokenType::END_ELEMENT, PIPELINE_TAG );
}

void MeasurementProvisionerConfigurationXml::parseMeasurementConfiguration ( MeasurementProvisionerConfiguration &, std::deque < sax::Token >::iterator & iter ) {
	FromXMLParserHelper::popToken ( iter, Token::TokenType::START_ELEMENT, MEASUREMENT_CONFIGURATION_TAG );

	 // skip everything for now
	while ( !FromXMLParserHelper::isTokenType ( iter, Token::TokenType::END_ELEMENT ) || FromXMLParserHelper::getTokenData ( iter, Token::TokenType::END_ELEMENT ) != MEASUREMENT_CONFIGURATION_TAG ) ++iter;

	FromXMLParserHelper::popToken ( iter, Token::TokenType::END_ELEMENT, MEASUREMENT_CONFIGURATION_TAG );
}

string MeasurementProvisionerConfigurationXml::MEASUREMENT_PROVISIONER_TAG = "MeasurementProvisioner";

string MeasurementProvisionerConfigurationXml::BINARIES_DIRECTORY_TAG = "BinariesDirectory";

string MeasurementProvisionerConfigurationXml::MEASUREMENT_PROVISION_TAG = "MeasurementProvision";

string MeasurementProvisionerConfigurationXml::INPUT_DATA_TAG = "InputData";
string MeasurementProvisionerConfigurationXml::INPUT_DATA_BATCH_TAG = "InputBatch";
string MeasurementProvisionerConfigurationXml::INPUT_DATA_BATCH_ATTR_DEPENDENCY_NAME = "dependency";

string MeasurementProvisionerConfigurationXml::INPUT_DATA_FILE_TAG = "File";
string MeasurementProvisionerConfigurationXml::INPUT_DATA_GENERATOR_TAG = "Generator";

string MeasurementProvisionerConfigurationXml::INPUT_DATA_ATTR_ID_NAME = "id";
string MeasurementProvisionerConfigurationXml::INPUT_DATA_ATTR_COUNT_NAME = "count";

string MeasurementProvisionerConfigurationXml::PIPELINES_TAG = "Pipelines";
string MeasurementProvisionerConfigurationXml::PIPELINE_TAG  = "Pipeline";
string MeasurementProvisionerConfigurationXml::COMMAND_TAG = "Command";

string MeasurementProvisionerConfigurationXml::MEASUREMENT_CONFIGURATION_TAG = "MeasurementConfiguration";
}
