/*
 * Author: Radovan Cerveny
 */

#include <cctype>
#include <unistd.h>
#include "MeasurementProvisionerInputData.hpp"
#include "MeasurementProvisionerUtils.hpp"
#include "exception/CommonException.h"

using namespace std;

namespace provisioner {

MPInputDatum::MPInputDatum ( MPInputDatum::Type type, std::string filename, std::string alias ) : type ( type ), filename ( std::move ( filename ) ), alias ( std::move ( alias ) ) {
}

MPInputDatum::~MPInputDatum ( ) {
	if ( type == MPInputDatum::Type::GENERATOR )
		unlink ( filename.c_str ( ) );
}

/*---------------------------------------------------------------------------------------------------------------*/

MPSubstitution::MPSubstitution ( const string & str ) {
	size_t prevPos = 0;

	for ( ; ; ) {
		size_t pos = findNextSubstitution ( str, prevPos );

		if ( pos != string::npos ) {

			tokens.emplace_back ( str.substr ( prevPos, pos - prevPos ) );
			tokens.emplace_back ( "" );

			 // remember the substitution id
			substitutionIdxs.push_back ( str[pos + 1] - '0' );

			 // adjust pos for $X
			pos += 2;
		} else {
			 // is there a remaining token?
			if ( prevPos != str.size ( ) )
				tokens.emplace_back ( str.substr ( prevPos, string::npos ) );

			break;
		}

		prevPos = pos;
	}

	substitutionPattern.insert ( substitutionIdxs.begin ( ), substitutionIdxs.end ( ) );
}

string MPSubstitution::substitute ( const MPSubstitutionMap & substitutionMap ) const {
	string res;
	int curSubstitutionIdx = 0;

	for ( const string & token : tokens ) {
		if ( token.size ( ) ) {
			res += token;
		} else {
			MPSubstitutionMap::const_iterator substitutionMapIter = substitutionMap.find ( substitutionIdxs[curSubstitutionIdx++] );

			if ( substitutionMapIter != substitutionMap.cend ( ) )
				res += substitutionMapIter->second;
		}
	}

	return res;
}

const set < int > & MPSubstitution::getSubstitutionPattern ( ) const {
	return substitutionPattern;
}

size_t MPSubstitution::findNextSubstitution ( const std::string & str, size_t prevPos ) {
	size_t pos = str.find_first_of ( SUBSTITUTION_DELIM, prevPos );

	if ( ( pos != string::npos ) && ( pos + 1 < str.size ( ) ) && isdigit ( str[pos + 1] ) )
		return pos;

	return string::npos;
}

set < int > MPSubstitution::requiredSubstitutionPattern ( const string & str ) {
	set < int > pattern;

	size_t prevPos = 0;
	size_t pos;

	while ( ( pos = findNextSubstitution ( str, prevPos ) ) != string::npos ) {
		pattern.insert ( str[pos + 1] - '0' );
		prevPos = pos + 2;
	}

	return pattern;
}

bool MPSubstitution::containsSubstitution ( const string & str ) {
	return findNextSubstitution ( str, 0 ) != string::npos;
}

/*---------------------------------------------------------------------------------------------------------------*/

MPPipelineCommand::MPPipelineCommand ( std::string str ) : MPSubstitution ( str ), rawCommand ( std::move ( str ) ) {
}

const string & MPPipelineCommand::getRawCommand ( ) const {
	return rawCommand;
}

/*---------------------------------------------------------------------------------------------------------------*/

MPInputData::MPInputData ( const MeasurementProvisionerConfiguration & mpc ) {
	processMPCInputData ( mpc );
}

const MPInputDatum & MPInputData::getMPIDHandle ( const string & filename ) const {
	if ( this->mpidHandles.find ( filename ) == this->mpidHandles.end ( ) )
		throw::exception::CommonException ( "MPInputData: trying to retrieve nonexistent handle: " + filename );

	return this->mpidHandles.find ( filename )->second;
}

const list < MPSubstitutionMap > & MPInputData::getSubstitutionMaps ( ) const {
	return this->substitutionMaps;
}

const list < MPPipeline > & MPInputData::getSubstitutionPipelines ( ) const {
	return this->substitutionPipelines;
}

void MPInputData::processMPCInputData ( const MeasurementProvisionerConfiguration & mpc ) {

	 // get requried substitution pattern and ensure every command has the same pattern
	for ( const MPCPipeline & pipeline : mpc.pipelines ) {
		MPPipeline mpp;

		for ( const string & command : pipeline ) {
			MPPipelineCommand mppc ( command );

			if ( this->requiredSubstitutionPattern.size ( ) == 0 )
				this->requiredSubstitutionPattern = mppc.getSubstitutionPattern ( );
			else if ( mppc.getSubstitutionPattern ( ).size ( ) && ( this->requiredSubstitutionPattern != mppc.getSubstitutionPattern ( ) ) )
				throw::exception::CommonException ( "MPInputData: Substitution pattern is not consistent across commands" );

			mpp.push_back ( std::move ( mppc ) );
		}

		substitutionPipelines.push_back ( std::move ( mpp ) );
	}

	// process input batches and generated substitution maps, throw if there is any discrepancy in substitution patterns or errors during data generation

	for ( const MPCInputBatch & inputBatch : mpc.inputBatches ) {

		 // generate substitution pattern in the batch, determine whether it is simple or dependency batch
		bool isSimpleBatch = !inputBatch.forcedDependencyBatch;
		set < int > inputSubstitutionPattern;

		for ( const MPCInputDatum & inputDatum : inputBatch.batch ) {
			if ( MPSubstitution::containsSubstitution ( inputDatum.value ) )
				isSimpleBatch = false;

			inputSubstitutionPattern.insert ( inputDatum.id );
		}

		if ( inputSubstitutionPattern != this->requiredSubstitutionPattern )
			throw::exception::CommonException ( "MPInputData: Substitution pattern in input data is not consistent with commands" );

		if ( isSimpleBatch )
			processSimpleBatch ( inputBatch );
		else
			processDependencyBatch ( inputBatch );
	}
}

void MPInputData::insertMPInputDatum ( MPInputDatum::Type type, std::string filename, std::string alias ) {
	 // if we already have a handle on the file, just return
	if ( this->mpidHandles.find ( filename ) != this->mpidHandles.end ( ) )
		return;

	MPInputDatum mpid ( type, std::move ( filename ), std::move ( alias ) );
	this->mpidHandles.emplace ( mpid.filename, std::move ( mpid ) );
}

void MPInputData::processSimpleBatchFileData ( const MPCInputDatum & inputDatum, list < string > & mpidList ) {
	list < string > expandedFilename = MPUtils::shellExpand ( inputDatum.value );

	for ( int i = 0; i < inputDatum.count; ++i )
		for ( const string & file : expandedFilename ) {
			insertMPInputDatum ( MPInputDatum::Type::FILE, file, file );
			mpidList.push_back ( file );
		}

}

void MPInputData::processSimpleBatchGeneratorData ( const MPCInputDatum & inputDatum, list < string > & mpidList ) {
	for ( int i = 0; i < inputDatum.count; ++i ) {
		string filename = MPUtils::generateTmpfileFromCommand ( inputDatum.value );
		insertMPInputDatum ( MPInputDatum::Type::GENERATOR, filename, inputDatum.value );
		mpidList.push_back ( std::move ( filename ) );
	}
}

void MPInputData::processSimpleBatch ( const MPCInputBatch & inputBatch ) {

	map < int, list < string > > distributedMPInputData;

	for ( const MPCInputDatum & inputDatum : inputBatch.batch ) {
		list < string > & didList = distributedMPInputData[inputDatum.id];

		switch ( inputDatum.type ) {
		case MPCInputDatum::Type::FILE:
			processSimpleBatchFileData ( inputDatum, didList );
			break;

		case MPCInputDatum::Type::GENERATOR:
			processSimpleBatchGeneratorData ( inputDatum, didList );
			break;
		}
	}

	 // check if we have enough data for each id to make substitution maps
	size_t didLen = distributedMPInputData.begin ( )->second.size ( );

	for ( const auto & didPair : distributedMPInputData )
		if ( didLen != didPair.second.size ( ) )
			throw::exception::CommonException ( "MPInputData: Not enough data for id " + to_string ( didPair.first ) + " in input batch to make substitution maps" );

	for ( size_t i = 0; i < didLen; ++i ) {
		MPSubstitutionMap substitutionMap;

		for ( auto & didPair : distributedMPInputData ) {
			substitutionMap[didPair.first] = std::move ( didPair.second.front ( ) );
			didPair.second.pop_front ( );
		}

		this->substitutionMaps.push_back ( std::move ( substitutionMap ) );
	}
}

void MPInputData::processDependencyBatchFileData ( const MPCInputDatum & inputDatum, const MPSubstitutionMap & prevSubstitutionMap, std::list < MPSubstitutionMap > & curSubstitutionMaps ) {
	list < string > expandedFilename = MPUtils::shellExpand ( inputDatum.value );

	for ( string & file : expandedFilename ) {
		insertMPInputDatum ( MPInputDatum::Type::FILE, file, file );

		for ( int i = 0; i < inputDatum.count; ++i ) {
			MPSubstitutionMap resultSubstitutionMap = prevSubstitutionMap;
			resultSubstitutionMap[inputDatum.id] = file;
			curSubstitutionMaps.push_back ( std::move ( resultSubstitutionMap ) );
		}
	}
}

void MPInputData::processDependencyBatchGeneratorData ( const MPCInputDatum & inputDatum, const MPSubstitutionMap & prevSubstitutionMap, std::list < MPSubstitutionMap > & curSubstitutionMaps ) {
	for ( int i = 0; i < inputDatum.count; ++i ) {

		MPSubstitutionMap resultSubstitutionMap = prevSubstitutionMap;
		MPSubstitution generatorSubstitution ( inputDatum.value );

		string filename = MPUtils::generateTmpfileFromCommand ( generatorSubstitution.substitute ( prevSubstitutionMap ) );
		insertMPInputDatum ( MPInputDatum::Type::GENERATOR, filename, inputDatum.value );

		resultSubstitutionMap[inputDatum.id] = filename;

		curSubstitutionMaps.push_back ( std::move ( resultSubstitutionMap ) );
	}
}

void MPInputData::processDependencyBatch ( const MPCInputBatch & inputBatch ) {
	// if the batch contains atleast one substitution, it is treated as dependency batch
	// thus we process InputData in order of ids from 1 to N

	 // distribute MPCInputData by their ids for processing in order
	map < int, list < MPCInputDatum > > distributedMPCInputData;

	for ( const MPCInputDatum & inputDatum : inputBatch.batch )
		distributedMPCInputData[inputDatum.id].push_back ( inputDatum );

	list < MPSubstitutionMap > prevSubstitutionMaps;

	for ( auto & didPair : distributedMPCInputData ) {
		list < MPSubstitutionMap > curSubstitutionMaps;

		 // treat first id data like simple batch data, because we forbid them to have substitutions
		if ( didPair.first == distributedMPCInputData.begin ( )->first ) {
			list < string > filenameList;

			for ( MPCInputDatum & inputDatum : didPair.second )
				switch ( inputDatum.type ) {
				case MPCInputDatum::Type::FILE:
					processSimpleBatchFileData ( inputDatum, filenameList );
					break;

				case MPCInputDatum::Type::GENERATOR:
					processSimpleBatchGeneratorData ( inputDatum, filenameList );
					break;
				}



			for ( auto & filename : filenameList ) {
				curSubstitutionMaps.push_back ( { { didPair.first, std::move ( filename ) }
												} );
			}
		} else {
			for ( MPCInputDatum & inputDatum : didPair.second ) {
				for ( const auto & prevSubstitutionMap : prevSubstitutionMaps )
					switch ( inputDatum.type ) {
					case MPCInputDatum::Type::FILE:
						processDependencyBatchFileData ( inputDatum, prevSubstitutionMap, curSubstitutionMaps );
						break;

					case MPCInputDatum::Type::GENERATOR:
						processDependencyBatchGeneratorData ( inputDatum, prevSubstitutionMap, curSubstitutionMaps );
						break;
					}

			}
		}

		swap ( prevSubstitutionMaps, curSubstitutionMaps );
	}

	this->substitutionMaps.splice ( this->substitutionMaps.end ( ), prevSubstitutionMaps );
}

}
