/*
 * Author: Radovan Cerveny
 */

#ifndef MEASUREMENT_PROVISIONER_INPUT_DATA_HPP_
#define MEASUREMENT_PROVISIONER_INPUT_DATA_HPP_

#include <map>
#include <set>
#include <vector>
#include <list>
#include <string>
#include "MeasurementProvisionerConfiguration.hpp"

namespace provisioner {

struct MPInputDatum {
	enum class Type {
		FILE, GENERATOR
	};

	Type		type;
	std::string filename;
	std::string alias;

	MPInputDatum ( const MPInputDatum & ) = delete;
	MPInputDatum & operator =( const MPInputDatum & ) = delete;
	MPInputDatum ( MPInputDatum && ) = default;
	MPInputDatum & operator =( MPInputDatum && ) = default;

	MPInputDatum ( Type, std::string, std::string );
	~MPInputDatum ( );
};

using MPSubstitutionMap = std::map < int, std::string >;

class MPSubstitution {
	std::vector < std::string > tokens;
	std::vector < int > substitutionIdxs;
	std::set < int > substitutionPattern;

public:
	MPSubstitution ( const std::string & );
	std::string substitute ( const MPSubstitutionMap & ) const;
	const std::set < int > & getSubstitutionPattern ( ) const;

private:
	static size_t findNextSubstitution ( const std::string &, size_t );

public:
	static std::set < int > requiredSubstitutionPattern ( const std::string & );
	static bool containsSubstitution ( const std::string & );

	static const char SUBSTITUTION_DELIM = '$';
};

class MPPipelineCommand : public MPSubstitution {
	std::string rawCommand;

public:
	MPPipelineCommand ( std::string );
	const std::string & getRawCommand ( ) const;
};

using MPPipeline = std::list < MPPipelineCommand >;

class MPInputData {
	std::map < std::string, MPInputDatum > mpidHandles;
	std::list < MPSubstitutionMap > substitutionMaps;
	std::set < int > requiredSubstitutionPattern;

	std::list < MPPipeline > substitutionPipelines;

	void insertMPInputDatum ( MPInputDatum::Type, std::string, std::string );

public:
	const MPInputDatum & getMPIDHandle ( const std::string & ) const;
	const std::list < MPSubstitutionMap > & getSubstitutionMaps ( ) const;
	const std::list < MPPipeline > & getSubstitutionPipelines ( ) const;
	MPInputData ( const MeasurementProvisionerConfiguration & );

private:
	void processSimpleBatchFileData ( const MPCInputDatum &, std::list < std::string > & );
	void processSimpleBatchGeneratorData ( const MPCInputDatum &, std::list < std::string > & );
	void processSimpleBatch ( const MPCInputBatch & );
	void processDependencyBatchFileData ( const MPCInputDatum &, const MPSubstitutionMap &, std::list < MPSubstitutionMap > & );
	void processDependencyBatchGeneratorData ( const MPCInputDatum &, const MPSubstitutionMap &, std::list < MPSubstitutionMap > & );
	void processDependencyBatch ( const MPCInputBatch & );
	void processMPCInputData ( const MeasurementProvisionerConfiguration & );
};

}

#endif /* MEASUREMENT_PROVISIONER_INPUT_DATA_HPP_ */
